import copy
import json
import logging
from typing import Any, Dict, List, Optional, Tuple, Type

import yaml  # type: ignore
from pydantic import BaseModel, Field, HttpUrl
from pydantic.fields import Undefined

from pait import field as pait_field
from pait.api_doc.base_parse import PaitBaseParse
from pait.g import config
from pait.model.core import PaitCoreModel
from pait.model.response import PaitBaseResponseModel
from pait.model.status import PaitStatus
from pait.util import create_pydantic_model, get_model_global_name, pait_model_schema

from .base_parse import FieldDictType, FieldSchemaTypeDict

__all__ = ["PaitOpenApi"]


class _OpenApiInfoModel(BaseModel):
    """open api info column model"""

    class _Contact(BaseModel):
        name: str
        url: str
        email: str

    class _License(BaseModel):
        name: str
        url: str

    title: str = Field("Pait Open Api")
    description: str = Field(None)
    version: str = Field("0.0.1")
    contact: _Contact = Field(None)
    license: _License = Field(None)


class _OpenApiTagModel(BaseModel):
    """openapi tag column model"""

    class _ExternalDocs(BaseModel):
        url: HttpUrl

    name: str
    description: str = Field(None)
    externalDocs: str = Field(None)


class _OpenApiServerModel(BaseModel):
    """openapi server column model"""

    url: str = Field("http://127.0.0.1")
    description: str = Field(None)


class PaitOpenApi(PaitBaseParse):
    def __init__(
        self,
        pait_dict: Dict[str, PaitCoreModel],
        title: Optional[str] = None,
        open_api_info: Optional[Dict[str, Any]] = None,
        open_api_tag_list: Optional[List[Dict[str, Any]]] = None,
        open_api_server_list: Optional[List[Dict[str, Any]]] = None,
        # default_response: Optional[...] = None,  # TODO
        type_: str = "json",
    ):
        super().__init__(pait_dict)
        self._header_keyword_dict: Dict[str, str] = {
            "Content-Type": "requestBody.content.<media-type>",
            "Accept": "responses.<code>.content.<media-type>",
            "Authorization": " security",
        }

        if not open_api_info:
            open_api_info = _OpenApiInfoModel(title=title).dict(exclude_none=True)
        else:
            _OpenApiInfoModel(**open_api_info)

        if not open_api_server_list:
            open_api_server_list = [_OpenApiServerModel().dict(exclude_none=True)]
        else:
            temp_open_api_server_list: List[Dict[str, Any]] = []
            for open_api_server in open_api_server_list:
                _OpenApiServerModel(**open_api_server)
                temp_open_api_server_list.append(open_api_server)

            open_api_server_list = temp_open_api_server_list

        if not open_api_tag_list:
            open_api_tag_list = []
        else:
            temp_open_api_tag_list: List[Dict[str, Any]] = []
            for open_api_tag in open_api_tag_list:
                _OpenApiTagModel(**open_api_tag)
                temp_open_api_tag_list.append(open_api_tag)
            open_api_tag_list = temp_open_api_tag_list

        self.open_api_dict: Dict[str, Any] = {
            "openapi": "3.0.0",
            "info": open_api_info,
            "servers": open_api_server_list,
            "tags": open_api_tag_list,
            "paths": {},
            "components": {"schemas": {}},
            # TODO
            # "security": {},
            # "externalDocs": {}
        }
        self.parse_data_2_openapi()
        if type_ == "json":
            self.content = json.dumps(self.open_api_dict, cls=config.json_encoder)
            self._content_type = ".json"
        elif type_ == "yaml":
            self.content = yaml.dump(self.open_api_dict, sort_keys=False)
            self._content_type = ".yaml"

    def _replace_pydantic_definitions(self, schema: dict, parent_schema: Optional[dict] = None) -> None:
        """update schemas'definitions to components schemas"""
        if not parent_schema:
            parent_schema = schema
        for key, value in schema.items():
            if key == "$ref" and not value.startswith("#/components"):
                index: int = value.rfind("/") + 1
                model_key: str = value[index:]
                schema[key] = f"#/components/schemas/{model_key}"
                self.open_api_dict["components"]["schemas"][model_key] = parent_schema["definitions"][model_key]
            elif isinstance(value, dict):
                self._replace_pydantic_definitions(value, parent_schema)
            elif isinstance(value, list):
                for item in value:
                    if isinstance(item, dict):
                        self._replace_pydantic_definitions(item, parent_schema)

    @staticmethod
    def _snake_name_to_hump_name(name: str) -> str:
        return "".join([item.capitalize() for item in name.split("_")])

    def _pait_model_2_response(
        self,
        pait_model: PaitCoreModel,
        openapi_method_dict: dict,
    ) -> None:
        openapi_response_dict: dict = openapi_method_dict.setdefault("responses", {})
        response_schema_dict: Dict[tuple, List[Dict[str, str]]] = {}
        for resp_model_class in pait_model.response_model_list:
            resp_model: PaitBaseResponseModel = resp_model_class()
            global_model_name: str = ""
            if isinstance(resp_model.response_data, type) and issubclass(resp_model.response_data, BaseModel):
                global_model_name = get_model_global_name(resp_model.response_data)

                schema_dict: dict = copy.deepcopy(pait_model_schema(resp_model.response_data))
                self._replace_pydantic_definitions(schema_dict)
                if "definitions" in schema_dict:
                    # fix del schema dict
                    del schema_dict["definitions"]
                self.open_api_dict["components"]["schemas"].update({global_model_name: schema_dict})

            for _status_code in resp_model.status_code:
                key: tuple = (_status_code, resp_model.media_type)
                if _status_code in openapi_response_dict:
                    if resp_model.description:
                        openapi_response_dict[_status_code]["description"] += f"|{resp_model.description}"
                    if resp_model.header:
                        openapi_response_dict[_status_code]["headers"].update(resp_model.header)
                else:
                    openapi_response_dict[_status_code] = {
                        "description": resp_model.description or "",
                        "headers": resp_model.header,
                    }
                if _status_code == 204:
                    # 204 No Content, have no body.
                    # To indicate the response body is empty, do not specify a content for the response
                    continue

                openapi_response_dict[_status_code]["content"] = {}
                if resp_model.links_schema_dict:
                    openapi_response_dict[_status_code]["links"] = resp_model.links_schema_dict

                if global_model_name:
                    openapi_schema_dict: dict = {"$ref": f"#/components/schemas/{global_model_name}"}
                    if key in response_schema_dict:
                        response_schema_dict[key].append(openapi_schema_dict)
                    else:
                        response_schema_dict[key] = [openapi_schema_dict]
                elif resp_model.openapi_schema:
                    if resp_model.media_type in openapi_response_dict[_status_code]["content"]:
                        raise ValueError(
                            f"{resp_model.media_type} already exists, "
                            f"Please check {pait_model.operation_id}'s "
                            f"response model list:{pait_model.response_model_list}"
                        )
                    openapi_response_dict[_status_code]["content"][resp_model.media_type] = resp_model.openapi_schema
                else:
                    logging.warning(
                        f"Can not found response schema from {pait_model.operation_id}'s response model:{resp_model}"
                    )

        # mutli response support
        # only response example see https://swagger.io/docs/specification/describing-responses/   FAQ
        for key_tuple, path_list in response_schema_dict.items():
            status_code, media_type = key_tuple
            if len(path_list) == 1:
                openapi_schema_dict = path_list[0]
            else:
                openapi_schema_dict = {"oneOf": path_list}
            openapi_response_dict[status_code]["content"] = {media_type: {"schema": openapi_schema_dict}}

    @staticmethod
    def _field_list_2_request_parameter(
        field_class: Type[pait_field.BaseField],
        openapi_method_dict: dict,
        field_dict_list: List[FieldSchemaTypeDict],
    ) -> None:
        openapi_parameters_list: list = openapi_method_dict.setdefault("parameters", [])
        for field_dict in field_dict_list:
            param_name: str = field_dict["raw"]["param_name"]
            required: bool = field_dict["default"] is Undefined
            # openapi description must not null, but Field().description value is None
            description: str = field_dict["description"] or ""
            if field_class == pait_field.Header:
                pass
                # I don't know why the swagger doc tells me to do a title name conversion
                #  when I don't actually need one
                # See: Header Parameters in
                #   https://swagger.io/docs/specification/describing-parameters/

                # param_name = self._header_keyword_dict.get(param_name, param_name)
            elif field_class == pait_field.Path and not required:
                raise ValueError("That path parameters must have required: true, " "because they are always required")
            elif field_class == pait_field.Cookie:
                if field_dict["raw"]["field"].raw_return:
                    # fix swagger ui cookie type when pait_field.raw_return is True
                    field_dict["raw"]["schema"]["type"] = "string"
                description += (
                    " "
                    "\n"
                    ">Note for Swagger UI and Swagger Editor users: "
                    " "
                    "\n"
                    ">Cookie authentication is"
                    'currently not supported for "try it out" requests due to browser security'
                    "restrictions. "
                    "See [this issue](https://github.com/swagger-api/swagger-js/issues/1163)"
                    "for more information. "
                    "[SwaggerHub](https://swagger.io/tools/swaggerhub/)"
                    "does not have this limitation. "
                )

            # When the required parameter is False and schema.default is empty,
            # openapi will automatically read the value of schema.example
            openapi_parameters_list.append(
                {
                    "name": param_name,
                    "in": field_class.get_field_name(),
                    "required": required,
                    "description": description,
                    "schema": field_dict["raw"]["schema"],
                }
            )

    @staticmethod
    def _field_list_2_file_upload(
        field_class: Type[pait_field.BaseField],
        openapi_method_dict: dict,
        field_dict_list: List[FieldSchemaTypeDict],
    ) -> None:
        """https://swagger.io/docs/specification/describing-request-body/file-upload/"""
        openapi_request_body_dict: dict = openapi_method_dict.setdefault("requestBody", {"content": {}})
        required_column_list: List[str] = [
            field_dict["raw"]["param_name"] for field_dict in field_dict_list if field_dict["default"] is Undefined
        ]
        properties_dict: dict = {
            field_dict["raw"]["param_name"]: {
                "title": field_dict["raw"]["param_name"].capitalize(),
                "type": "string",
                "format": "binary",
            }
            for field_dict in field_dict_list
        }
        if not openapi_request_body_dict.get("required", False):
            if required_column_list:
                openapi_request_body_dict["required"] = True
        if field_class.media_type not in openapi_request_body_dict["content"]:
            openapi_request_body_dict["content"][field_class.media_type] = {
                "schema": {"type": "object", "properties": properties_dict, "required": required_column_list}
            }
        else:
            openapi_request_body_dict["content"][field_class.media_type]["schema"]["properties"].update(properties_dict)
            openapi_request_body_dict["content"][field_class.media_type]["schema"]["required"].extend(
                required_column_list
            )

    def _field_list_2_request_body(
        self,
        field_class: Type[pait_field.BaseField],
        openapi_method_dict: dict,
        field_dict_list: List[FieldSchemaTypeDict],
        operation_id: str,
    ) -> None:
        """
        gen request body schema and update request body schemas'definitions to components schemas
        Doc: https://swagger.io/docs/specification/describing-request-body/
        """
        openapi_request_body_dict: dict = openapi_method_dict.setdefault("requestBody", {"content": {}})
        annotation_dict: Dict[str, Tuple[Type, pait_field.BaseField]] = {
            field_dict["raw"]["param_name"]: (field_dict["raw"]["annotation"], field_dict["raw"]["field"])
            for field_dict in field_dict_list
        }
        _pydantic_model: Type[BaseModel] = create_pydantic_model(
            annotation_dict, class_name=f"{self._snake_name_to_hump_name(operation_id)}DynamicModel"
        )
        schema_dict = copy.deepcopy(pait_model_schema(_pydantic_model))
        # pait will disassemble and reassemble the BaseModel, so there is no way to reuse the model in openapi.
        # TODO support model
        # self._replace_pydantic_definitions(schema_dict)
        if "definitions" in schema_dict:
            del schema_dict["definitions"]
        if field_class.media_type in openapi_request_body_dict["content"]:
            for key, value in openapi_request_body_dict["content"][field_class.media_type]["schema"].items():
                if isinstance(value, list):
                    value.extend(schema_dict[key])
                elif isinstance(value, dict):
                    value.update(schema_dict[key])
        else:
            openapi_request_body_dict["content"][field_class.media_type] = {"schema": schema_dict}

        if field_class == pait_field.MultiForm:
            # pait_field.File must precede pait_field.MultiForm
            form_encoding_dict = openapi_request_body_dict["content"][field_class.media_type].setdefault("encoding", {})
            for field_dict in field_dict_list:
                if pait_field.File.media_type in openapi_request_body_dict["content"]:
                    openapi_request_body_dict["content"][field_class.media_type]["schema"]["properties"][
                        field_dict["raw"]["param_name"]
                    ]["description"] += (
                        " " " " "\n" f">Swagger UI could not support, when media_type is {pait_field.File.media_type}"
                    )
                form_encoding_dict[field_dict["raw"]["param_name"]] = field_class.openapi_serialization
            # TODO support payload?
            # https://swagger.io/docs/specification/describing-request-body/

    # flake8: noqa: C901
    def parse_data_2_openapi(self) -> None:
        for group, pait_model_list in self._group_pait_dict.items():
            for pait_model in pait_model_list:
                path: str = pait_model.openapi_path
                openapi_path_dict: dict = self.open_api_dict["paths"].setdefault(path, {})
                for method in pait_model.method_list:
                    # Documentation for each route
                    openapi_method_dict: dict = openapi_path_dict.setdefault(method.lower(), {})
                    openapi_method_dict["group"] = group  # Not an openapi standard parameter
                    if pait_model.tag:
                        openapi_method_dict["tags"] = list(pait_model.tag)
                        # Additional labeling instructions
                        for tag in pait_model.tag:
                            if tag not in {tag_dict["name"] for tag_dict in self.open_api_dict["tags"]}:
                                logging.warning(f"Can not found tag:{tag} description, set default description")
                                self.open_api_dict["tags"].append(
                                    {
                                        "name": tag,
                                        "description": "",
                                    }
                                )
                    if pait_model.status in (
                        PaitStatus.abnormal,
                        PaitStatus.maintenance,
                        PaitStatus.archive,
                        PaitStatus.abandoned,
                    ):
                        openapi_method_dict["deprecated"] = True
                    openapi_method_dict["summary"] = pait_model.summary
                    openapi_method_dict["description"] = pait_model.desc
                    openapi_method_dict["operationId"] = f"{method}.{pait_model.operation_id}"

                    all_field_dict: FieldDictType = self._parse_pait_model_to_field_dict(pait_model)
                    all_field_class_list: List[Type[pait_field.BaseField]] = sorted(
                        [i for i in all_field_dict.keys()], key=lambda x: x.get_field_name()
                    )
                    for field_class in all_field_class_list:
                        field_dict_list = all_field_dict[field_class]
                        if field_class in (
                            pait_field.Cookie,
                            pait_field.Header,
                            pait_field.Path,
                            pait_field.Query,
                            pait_field.MultiQuery,
                        ):
                            self._field_list_2_request_parameter(field_class, openapi_method_dict, field_dict_list)
                        elif field_class in (pait_field.Body, pait_field.Form, pait_field.MultiForm):
                            self._field_list_2_request_body(
                                field_class, openapi_method_dict, field_dict_list, pait_model.operation_id
                            )
                        elif field_class in (pait_field.File,):
                            self._field_list_2_file_upload(field_class, openapi_method_dict, field_dict_list)
                        else:
                            logging.warning(f"Pait not support field:{field_class}")

                    if pait_model.response_model_list:
                        self._pait_model_2_response(pait_model, openapi_method_dict)



import inspect
import json
import logging
from datetime import datetime
from decimal import Decimal
from enum import Enum
from json import JSONEncoder
from typing import Any, Dict, List, Optional, Tuple, Type, get_type_hints

from pydantic import BaseConfig, BaseModel, create_model

from pait.exceptions import PaitBaseException
from pait.field import BaseField, Depends, is_pait_field

from ._func_sig import FuncSig


class UndefinedType:
    def __repr__(self) -> str:
        return "PaitUndefined"


# pait undefined flag
Undefined: UndefinedType = UndefinedType()

json_type_default_value_dict: Dict[str, Any] = {
    "null": None,
    "bool": True,
    "boolean": True,
    "string": "",
    "number": 0.0,
    "float": 0.0,
    "integer": 0,
    "object": {},
    "array": [],
}

python_type_default_value_dict: Dict[type, Any] = {
    bool: True,
    float: 0.0,
    int: 0,
    str: "",
    list: [],
    tuple: [],
    datetime: 0,
    Decimal: 0,
}


def create_pydantic_model(
    annotation_dict: Dict[str, Tuple[Type, Any]],
    class_name: str = "DynamicModel",
    pydantic_config: Type[BaseConfig] = None,
    pydantic_base: Type[BaseModel] = None,
    pydantic_module: str = "pydantic.main",
    pydantic_validators: Dict[str, classmethod] = None,
) -> Type[BaseModel]:
    """pydantic create_model helper
    if use create_model('DynamicModel', **annotation_dict), mypy will tip error
    """
    return create_model(
        class_name,
        __config__=pydantic_config,
        __base__=pydantic_base,
        __module__=pydantic_module,
        __validators__=pydantic_validators,
        **annotation_dict,
    )


def gen_example_json_from_python(obj: Any) -> Any:
    if isinstance(obj, dict):
        new_dict: dict = {}
        for key, value in obj.items():
            new_dict[key] = gen_example_json_from_python(value)
        return new_dict
    else:
        return python_type_default_value_dict.get(type(obj), obj)


def gen_example_dict_from_schema(schema_dict: Dict[str, Any], definition_dict: Optional[dict] = None) -> Dict[str, Any]:
    gen_dict: Dict[str, Any] = {}
    property_dict: Dict[str, Any] = schema_dict["properties"]
    if not definition_dict:
        _definition_dict: dict = schema_dict.get("definitions", {})
    else:
        _definition_dict = definition_dict
    for key, value in property_dict.items():
        if "items" in value and value["type"] == "array":
            if "$ref" in value["items"]:
                model_key: str = value["items"]["$ref"].split("/")[-1]
                gen_dict[key] = [gen_example_dict_from_schema(_definition_dict.get(model_key, {}), _definition_dict)]
            else:
                gen_dict[key] = []
        elif "$ref" in value:
            model_key = value["$ref"].split("/")[-1]
            gen_dict[key] = gen_example_dict_from_schema(_definition_dict.get(model_key, {}), _definition_dict)
        else:
            if "example" in value:
                gen_dict[key] = value["example"]
                if inspect.isfunction(gen_dict[key]):
                    gen_dict[key] = gen_dict[key]()
            elif "default" in value:
                gen_dict[key] = value["default"]
            else:
                if "type" in value:
                    if value["type"] not in json_type_default_value_dict:
                        raise KeyError(f"Can not found type: {key} in json type")
                    gen_dict[key] = json_type_default_value_dict[value["type"]]
                else:
                    gen_dict[key] = "object"
            if isinstance(gen_dict[key], Enum):
                gen_dict[key] = gen_dict[key].value
    return gen_dict


def gen_example_json_from_schema(schema_dict: Dict[str, Any], cls: Optional[Type[JSONEncoder]] = None) -> str:
    return json.dumps(gen_example_dict_from_schema(schema_dict), cls=cls)


def get_parameter_list_from_pydantic_basemodel(pait_model: Type[BaseModel]) -> List["inspect.Parameter"]:
    """get class parameter list by attributes, if attributes not default value, it will be set `Undefined`"""
    parameter_list: Optional[List["inspect.Parameter"]] = getattr(pait_model, "_parameter_list", None)
    if parameter_list is not None:
        return parameter_list
    parameter_list = []
    for key, model_field in pait_model.__fields__.items():
        if not is_pait_field(model_field.field_info):
            raise TypeError(f"{model_field.field_info} must instance {BaseField} or {Depends}")
        annotation: Type = pait_model.__annotations__[key]
        if getattr(annotation, "real", None):
            # support like pydantic.ConstrainedIntValue
            annotation = annotation.real.__objclass__  # type: ignore
        parameter = inspect.Parameter(
            key,
            inspect.Parameter.POSITIONAL_ONLY,
            default=model_field.field_info,
            annotation=annotation,
        )
        parameter_list.append(parameter)

    setattr(pait_model, "_parameter_list", parameter_list)
    return parameter_list


def get_parameter_list_from_class(cbv_class: Type) -> List["inspect.Parameter"]:
    """get class parameter list by attributes, if attributes not default value, it will be set `Undefined`"""
    parameter_list: Optional[List["inspect.Parameter"]] = getattr(cbv_class, "_parameter_list", None)
    if parameter_list is not None:
        return parameter_list
    parameter_list = []
    if hasattr(cbv_class, "__annotations__"):
        for param_name, param_annotation in get_type_hints(cbv_class).items():
            default: Any = getattr(cbv_class, param_name, Undefined)
            if not is_pait_field(default):
                continue
            parameter: "inspect.Parameter" = inspect.Parameter(
                param_name,
                inspect.Parameter.POSITIONAL_ONLY,
                default=default,
                annotation=param_annotation,
            )
            parameter_list.append(parameter)
    setattr(cbv_class, "_parameter_list", parameter_list)
    return parameter_list


def gen_tip_exc(_object: Any, exception: "Exception", parameter: Optional[inspect.Parameter] = None) -> Exception:
    """Help users understand which parameter is wrong"""
    if not parameter and getattr(exception, "_is_tip_exc", None):
        return exception
    if parameter:
        param_value: BaseField = parameter.default
        annotation: Type[BaseModel] = parameter.annotation
        param_name: str = parameter.name

        parameter_value_name: str = param_value.__class__.__name__
        if param_value is parameter.empty:
            param_str: str = f"{param_name}: {annotation}"
        else:
            param_str = f"{param_name}: {annotation} = {parameter_value_name}"
            if isinstance(param_value, BaseField):
                param_str += f"(alias={param_value.alias}, default={param_value.default})"
    else:
        param_str = ""

    file: Optional[str] = None
    if isinstance(_object, FuncSig):
        _object = _object.func
    if inspect.isfunction(_object):
        title: str = "def"
        if inspect.iscoroutinefunction(_object):
            title = "async def"
        file = inspect.getfile(_object)
        line: int = inspect.getsourcelines(_object)[1]
        error_object_name: str = _object.__name__
        logging.debug(
            f"""
{title} {error_object_name}(
    ...
    {param_str} <-- error
    ...
):
    pass
"""
        )
    else:
        module: Any = inspect.getmodule(_object)
        if module:
            file = module.__file__
        line = inspect.getsourcelines(_object.__class__)[1]
        error_object_name = _object.__class__.__name__
        if "class" in error_object_name:
            error_object_name = str(_object.__class__)
        logging.debug(f"class: `{error_object_name}`  attributes error\n    {param_str}")
    exc_msg: str = f'File "{file}",' f" line {line}," f" in {error_object_name}." f" error:{str(exception)}"
    try:
        exc: Exception = exception.__class__(exc_msg)
    except Exception:
        exc = PaitBaseException(exc_msg)
    setattr(exception, "_is_tip_exc", True)
    return exc



import json
from tempfile import NamedTemporaryFile
from typing import Type

from flask import Response, jsonify, make_response, send_from_directory

from pait.app.base.app_helper import BaseAppHelper
from pait.core import Pait as _Pait
from pait.g import config
from pait.model import response

from ._app_helper import AppHelper

__all__ = ["pait"]


def make_mock_response(pait_response: Type[response.PaitBaseResponseModel]) -> Response:
    if issubclass(pait_response, response.PaitJsonResponseModel):
        resp: Response = jsonify(json.loads(pait_response.get_example_value(json_encoder_cls=config.json_encoder)))
    elif issubclass(pait_response, response.PaitTextResponseModel) or issubclass(
        pait_response, response.PaitHtmlResponseModel
    ):
        resp = make_response(pait_response.get_example_value())
        resp.mimetype = pait_response.media_type
    elif issubclass(pait_response, response.PaitFileResponseModel):
        with NamedTemporaryFile(delete=True) as temporary_file:
            temporary_file.write(pait_response.get_example_value())
            temporary_file.seek(0)
            _, f_path, f_filename = temporary_file.name.split("/")
            resp = send_from_directory("/" + f_path, f_filename, mimetype=pait_response.media_type)
    else:
        raise NotImplementedError(f"make_mock_response not support {pait_response}")
    resp.status_code = pait_response.status_code[0]
    if pait_response.header:
        resp.headers.update(pait_response.header)  # type: ignore
    return resp


class Pait(_Pait):
    app_helper_class: "Type[BaseAppHelper]" = AppHelper
    # If you assign a value directly, it will become a bound function
    make_mock_response_fn: staticmethod = staticmethod(make_mock_response)
    # make_mock_response_fn: Callable[[Type[response.PaitBaseResponseModel]], Any] = staticmethod(make_mock_response)


pait = Pait()



from typing import TYPE_CHECKING, Any, Dict, Optional, Set, Type, Union

from pydantic.schema import (
    default_ref_template,
    get_flat_models_from_model,
    get_long_model_name,
    get_model,
    get_schema_ref,
    model_process_schema,
    normalize_name,
)

if TYPE_CHECKING:
    from pydantic import BaseModel
    from pydantic.dataclasses import Dataclass
    from pydantic.schema import TypeModelOrEnum, TypeModelSet


global_name_model_map = {}
global_conflicting_names: Set[str] = set()


def get_model_global_name(model: "TypeModelOrEnum") -> str:
    return pait_get_model_name_map({model})[model]


def pait_get_model_name_map(unique_models: "TypeModelSet") -> Dict["TypeModelOrEnum", str]:
    """
    Process a set of models and generate unique names for them to be used as keys in the JSON Schema
    definitions. By default the names are the same as the class name. But if two models in different Python



    modules have the same name (e.g. "users.Model" and "items.Model"), the generated names will be
    based on the Python module path for those conflicting models to prevent name collisions.

    :param unique_models: a Python set of models
    :return: dict mapping models to names
    """
    global global_name_model_map
    global global_conflicting_names

    for model in unique_models:
        model_name = normalize_name(model.__name__)
        if model_name in global_conflicting_names:
            model_name = get_long_model_name(model)
            global_name_model_map[model_name] = model
        elif model_name in global_name_model_map:
            global_conflicting_names.add(model_name)
            conflicting_model = global_name_model_map.pop(model_name)
            global_name_model_map[get_long_model_name(conflicting_model)] = conflicting_model
            global_name_model_map[get_long_model_name(model)] = model
        else:
            global_name_model_map[model_name] = model
    return {v: k for k, v in global_name_model_map.items()}


def pait_model_schema(
    model: Union[Type["BaseModel"], Type["Dataclass"]],
    by_alias: bool = True,
    ref_prefix: Optional[str] = None,
    ref_template: str = default_ref_template,
) -> Dict[str, Any]:
    """
    Generate a JSON Schema for one model. With all the sub-models defined in the ``definitions`` top-level
    JSON key.

    :param model: a Pydantic model (a class that inherits from BaseModel)
    :param by_alias: generate the schemas using the aliases defined, if any
    :param ref_prefix: the JSON Pointer prefix for schema references with ``$ref``, if None, will be set to the
      default of ``#/definitions/``. Update it if you want the schemas to reference the definitions somewhere
      else, e.g. for OpenAPI use ``#/components/schemas/``. The resulting generated schemas will still be at the
      top-level key ``definitions``, so you can extract them from there. But all the references will have the set
      prefix.
    :param ref_template: Use a ``string.format()`` template for ``$ref`` instead of a prefix. This can be useful for
      references that cannot be represented by ``ref_prefix`` such as a definition stored in another file. For a
      sibling json file in a ``/schemas`` directory use ``"/schemas/${model}.json#"``.
    :return: dict with the JSON Schema for the passed ``model``
    """
    model = get_model(model)
    flat_models = get_flat_models_from_model(model)
    model_name_map = pait_get_model_name_map(flat_models)
    model_name = model_name_map[model]
    m_schema, m_definitions, nested_models = model_process_schema(
        model, by_alias=by_alias, model_name_map=model_name_map, ref_prefix=ref_prefix, ref_template=ref_template
    )
    if model_name in nested_models:
        # model_name is in Nested models, it has circular references
        m_definitions[model_name] = m_schema
        m_schema = get_schema_ref(model_name, ref_prefix, ref_template, False)
    if m_definitions:
        m_schema.update({"definitions": m_definitions})
    return m_schema



from ._func_sig import FuncSig, get_func_sig
from ._lazy_property import LazyProperty
from ._pydantic_util import get_model_global_name, pait_get_model_name_map, pait_model_schema
from ._util import (
    Undefined,
    UndefinedType,
    create_pydantic_model,
    gen_example_dict_from_schema,
    gen_example_json_from_python,
    gen_example_json_from_schema,
    gen_tip_exc,
    get_parameter_list_from_class,
    get_parameter_list_from_pydantic_basemodel,
    json_type_default_value_dict,
    python_type_default_value_dict,
)



import logging
from typing import Any, Dict, List, Optional

from flask import Blueprint, Flask, current_app, request
from werkzeug.exceptions import NotFound

from pait.api_doc.html import get_redoc_html as _get_redoc_html
from pait.api_doc.html import get_swagger_ui_html as _get_swagger_ui_html
from pait.api_doc.open_api import PaitOpenApi
from pait.field import Depends, Query
from pait.model.core import PaitCoreModel

from ._load_app import load_app
from ._pait import pait

__all__ = ["add_doc_route"]


def add_doc_route(
    app: Flask,
    scheme: Optional[str] = None,
    prefix: str = "/",
    pin_code: str = "",
    title: str = "Pait Doc",
    open_api_tag_list: Optional[List[Dict[str, Any]]] = None,
) -> None:
    if pin_code:
        logging.info(f"doc route start pin code:{pin_code}")

    def _get_request_pin_code(r_pin_code: str = Query.i("", alias="pin_code")) -> Optional[str]:
        if pin_code:
            if r_pin_code != pin_code:
                raise NotFound
        return r_pin_code

    def _get_open_json_url(r_pin_code: str) -> str:
        openapi_json_url: str = f"{'/'.join(request.path.split('/')[:-1])}/openapi.json"
        if r_pin_code:
            openapi_json_url += f"?pin_code={r_pin_code}"
        return openapi_json_url

    @pait()
    def get_redoc_html(r_pin_code: str = Depends.i(_get_request_pin_code)) -> str:
        return _get_redoc_html(_get_open_json_url(r_pin_code), title)

    @pait()
    def get_swagger_ui_html(r_pin_code: str = Depends.i(_get_request_pin_code)) -> str:
        return _get_swagger_ui_html(_get_open_json_url(r_pin_code), title)

    @pait(pre_depend_list=[_get_request_pin_code])
    def openapi_route() -> dict:
        pait_dict: Dict[str, PaitCoreModel] = load_app(current_app)
        _scheme: str = scheme or request.scheme
        pait_openapi: PaitOpenApi = PaitOpenApi(
            pait_dict,
            title=title,
            open_api_server_list=[{"url": f"{_scheme}://{request.host}", "description": ""}],
            open_api_tag_list=open_api_tag_list,
        )
        return pait_openapi.open_api_dict

    blueprint: Blueprint = Blueprint("api doc", __name__, url_prefix=prefix)
    blueprint.add_url_rule("/redoc", view_func=get_redoc_html, methods=["GET"])
    blueprint.add_url_rule("/swagger", view_func=get_swagger_ui_html, methods=["GET"])
    blueprint.add_url_rule("/openapi.json", view_func=openapi_route, methods=["GET"])
    app.register_blueprint(blueprint)



from typing import TYPE_CHECKING, List, Type

from pydantic import BaseModel

if TYPE_CHECKING:
    from pait import field

    from .core import PaitCoreModel
    from .response import PaitBaseResponseModel


class LinksModel(object):
    param_name: str
    pait_field: "field.BaseField"
    link_name: str
    operation_id: str
    parameters_dict: dict

    def __init__(self, pait_response_model: "Type[PaitBaseResponseModel]", openapi_runtime_expr: str, desc: str = ""):
        """
        doc: https://swagger.io/docs/specification/links/
        :param pait_response_model: pait response model
        :param openapi_runtime_expr: open api runtime expression syntax.
            Only support $response.headerXXX or $response.bodyXXX
            Please refer to the section of the document at `Runtime Expression Syntax`
        :param desc: links desc
        """
        self.openapi_runtime_expr: str = openapi_runtime_expr
        self.pait_response_model: "Type[PaitBaseResponseModel]" = pait_response_model
        self.desc: str = desc
        self._cache: dict = {}

    def _check_openapi_runtime_expr(self) -> None:
        if self.openapi_runtime_expr.startswith("$response.header."):
            header_key: str = self.openapi_runtime_expr.replace("$response.header.", "")
            if header_key not in self.pait_response_model.header:
                raise KeyError(f"Can not found header key:{header_key} from {self.pait_response_model}")
        elif self.openapi_runtime_expr.startswith("$response.body#"):
            if not self.pait_response_model.is_base_model_response_data():
                raise RuntimeError(
                    f"Expr: {self.openapi_runtime_expr} only support "
                    f"pait_response_model.response_data type is pydantic.Basemodel"
                )
            try:
                key_list: List[str] = [
                    i for i in self.openapi_runtime_expr.replace("$response.body#", "").split("/") if i
                ]
            except Exception:
                raise RuntimeError(
                    f"Check expr: {self.openapi_runtime_expr} error."
                    "Please refer to the section of the document(https://swagger.io/docs/specification/links/) "
                    "at `Runtime Expression Syntax`"
                )

            base_model: Type[BaseModel] = self.pait_response_model.response_data  # type: ignore
            for key in key_list:
                if key not in base_model.__fields__:
                    raise ValueError(
                        f"check expr:{self.openapi_runtime_expr} error "  # type: ignore
                        f"from {self.pait_response_model.response_data}"  # type: ignore
                    )
                temp_type: Type = base_model.__fields__[key].type_
                if issubclass(temp_type, BaseModel):
                    base_model = temp_type
        else:
            raise ValueError(
                "Only support $response.headerXXX or $response.bodyXXX. "
                "Please refer to the section of the document(https://swagger.io/docs/specification/links/) "
                "at `Runtime Expression Syntax`"
            )

    def register(self, pait_model: "PaitCoreModel", param_name: str, pait_field: "field.BaseField") -> None:
        """
        In order not to affect the startup speed, this method is only allowed to be called under the openapi module"""
        param_name = pait_field.alias or param_name
        for method in pait_model.method_list:
            operation_id: str = f"{method}.{pait_model.operation_id}"
            global_key: str = f"{operation_id}/{pait_field.get_field_name()}/{param_name}"
            if global_key in self._cache:
                return

            self._check_openapi_runtime_expr()
            self.pait_response_model.register_link_schema(
                {
                    global_key: {
                        "description": self.desc,
                        "operationId": operation_id,
                        "parameters": {param_name: self.openapi_runtime_expr},
                    }
                }
            )



from typing import Dict, Mapping, Optional

from flask import Response
from flask.testing import FlaskClient

from pait.app.base import BaseTestHelper
from pait.model.core import PaitCoreModel

from ._load_app import load_app

__all__ = ["FlaskTestHelper"]


class FlaskTestHelper(BaseTestHelper[Response]):
    client: FlaskClient

    def _app_init_field(self) -> None:
        if self.file_dict:
            if self.form_dict:
                self.form_dict.update(self.file_dict)
            else:
                self.form_dict = self.file_dict

        if self.cookie_dict:
            for key, value in self.cookie_dict.items():
                self.client.set_cookie("localhost", key, value)

    def _gen_pait_dict(self) -> Dict[str, PaitCoreModel]:
        return load_app(self.client.application)

    @staticmethod
    def _get_status_code(resp: Response) -> int:
        return resp.status_code

    @staticmethod
    def _get_content_type(resp: Response) -> str:
        return resp.mimetype or ""

    @staticmethod
    def _get_headers(resp: Response) -> Mapping:
        return resp.headers  # type: ignore

    @staticmethod
    def _get_json(resp: Response) -> dict:
        return resp.get_json()

    @staticmethod
    def _get_text(resp: Response) -> str:
        return resp.get_data().decode()

    @staticmethod
    def _get_bytes(resp: Response) -> bytes:
        return resp.get_data()

    def _replace_path(self, path_str: str) -> Optional[str]:
        if self.path_dict and path_str[0] == "<" and path_str[-1] == ">":
            return self.path_dict[path_str[1:-1]]
        return None

    def _real_request(self, method: str) -> Response:
        return self.client.open(
            self.path, data=self.form_dict, json=self.body_dict, headers=self.header_dict, method=method
        )



import json
from typing import Any, AsyncContextManager, Type

import aiofiles  # type: ignore
from sanic import response as sanic_response
from sanic.response import json as resp_json
from sanic_testing.testing import SanicTestClient, TestingResponse  # type: ignore

from pait.app.base import BaseAppHelper
from pait.core import Pait as _Pait
from pait.g import config
from pait.model import response

from ._app_helper import AppHelper

__all__ = ["pait"]


async def make_mock_response(pait_response: Type[response.PaitBaseResponseModel]) -> sanic_response.BaseHTTPResponse:
    if issubclass(pait_response, response.PaitJsonResponseModel):
        resp: sanic_response.BaseHTTPResponse = resp_json(
            json.loads(pait_response.get_example_value(json_encoder_cls=config.json_encoder))
        )
    elif issubclass(pait_response, response.PaitTextResponseModel) or issubclass(
        pait_response, response.PaitHtmlResponseModel
    ):
        resp = sanic_response.text(pait_response.get_example_value(), content_type=pait_response.media_type)
    elif issubclass(pait_response, response.PaitFileResponseModel):
        named_temporary_file: AsyncContextManager = aiofiles.tempfile.NamedTemporaryFile()  # type: ignore
        f: Any = await named_temporary_file.__aenter__()
        await f.write(pait_response.get_example_value())
        await f.seek(0)
        resp = await sanic_response.file_stream(f.name, mime_type=pait_response.media_type)

        raw_streaming_fn = resp.streaming_fn

        async def _streaming_fn(_response: sanic_response.BaseHTTPResponse) -> None:
            await raw_streaming_fn(_response)
            await named_temporary_file.__aexit__(None, None, None)

        resp.streaming_fn = _streaming_fn
    else:
        raise NotImplementedError(f"make_mock_response not support {pait_response}")
    resp.status = pait_response.status_code[0]
    if pait_response.header:
        resp.headers.update(pait_response.header)
    return resp


class Pait(_Pait):
    app_helper_class: "Type[BaseAppHelper]" = AppHelper
    # If you assign a value directly, it will become a bound function
    make_mock_response_fn: staticmethod = staticmethod(make_mock_response)


pait = Pait()



import binascii
import json
import os
from io import BytesIO
from typing import Dict, Mapping, Optional, Tuple

from tornado.testing import AsyncHTTPTestCase, HTTPResponse

from pait.app.base import BaseTestHelper
from pait.model.core import PaitCoreModel

from ._load_app import load_app

__all__ = ["TornadoTestHelper"]


class TornadoTestHelper(BaseTestHelper[HTTPResponse]):
    client: AsyncHTTPTestCase

    def _app_init_field(self) -> None:
        if self.cookie_dict:
            self.header_dict.update(self.cookie_dict)
        if "$?" in self.path:
            self.path = self.path.replace("$?", "?")
        elif self.path.endswith("$"):
            self.path = self.path[:-1]

    def _gen_pait_dict(self) -> Dict[str, PaitCoreModel]:
        return load_app(self.client.get_app())

    @staticmethod
    def _get_status_code(resp: HTTPResponse) -> int:
        return resp.code

    @staticmethod
    def _get_content_type(resp: HTTPResponse) -> str:
        return resp.headers["Content-Type"]

    @staticmethod
    def _get_text(resp: HTTPResponse) -> str:
        return resp.body.decode()

    @staticmethod
    def _get_bytes(resp: HTTPResponse) -> bytes:
        return resp.body

    @staticmethod
    def _get_json(resp: HTTPResponse) -> dict:
        return json.loads(resp.body.decode())

    @staticmethod
    def _get_headers(resp: HTTPResponse) -> Mapping:
        return resp.headers

    def _replace_path(self, path_str: str) -> Optional[str]:
        if self.path_dict:
            head_index, tail_index = -1, -1
            for index, i in enumerate(path_str):
                if i == "<":
                    head_index = index
                if i == ">":
                    tail_index = index
            if head_index != -1 or tail_index != -1:
                return self.path_dict[path_str[head_index + 1 : tail_index]]
        return None

    def _real_request(self, method: str) -> HTTPResponse:
        method = method.upper()
        if self.file_dict or self.form_dict:
            if method != "POST":
                raise RuntimeError("Must use method post")
            content_type, body = self.encode_multipart_formdata(data=self.form_dict, files=self.file_dict)
            headers: dict = self.header_dict.copy()
            headers.update({"Content-Type": content_type, "content-length": str(len(body))})
            return self.client.fetch(self.path, method="POST", headers=headers, body=body)
        body_bytes: Optional[bytes] = None
        if self.body_dict:
            body_bytes = json.dumps(self.body_dict).encode()
        return self.client.fetch(self.path, method=method, headers=self.header_dict, body=body_bytes)

    @staticmethod
    def choose_boundary() -> str:
        """
        Our embarrassingly-simple replacement for mimetools.choose_boundary.
        """
        boundary: bytes = binascii.hexlify(os.urandom(16))
        return boundary.decode("ascii")

    def encode_multipart_formdata(self, data: Optional[dict] = None, files: Optional[dict] = None) -> Tuple[str, bytes]:
        """
        fields is a sequence of (name, value) elements for regular form fields.
        files is a sequence of (name, filename, value) elements for data to be
        uploaded as files.
        Return (content_type, body) ready for httplib.HTTP instance
        """
        body: BytesIO = BytesIO()
        boundary: str = self.choose_boundary()
        if data:
            for key, value in data.items():
                body.write(("--%s\r\n" % boundary).encode(encoding="utf-8"))
                body.write(('Content-Disposition:form-data;name="%s"\r\n' % key).encode(encoding="utf-8"))
                body.write("\r\n".encode(encoding="utf-8"))
                if isinstance(value, int):
                    value = str(value)
                body.write(("%s\r\n" % value).encode(encoding="utf-8"))

        if files:
            for key, value in files.items():
                body.write(("--%s\r\n" % boundary).encode(encoding="utf-8"))
                body.write(
                    ('Content-Disposition:form-data;name="file";filename="%s"\r\n' % key).encode(encoding="utf-8")
                )
                body.write("\r\n".encode(encoding="utf-8"))
                body.write(value)
                body.write("\r\n".encode(encoding="utf-8"))

        body.write(("--%s--\r\n" % boundary).encode(encoding="utf-8"))
        content_type: str = "multipart/form-data;boundary=%s" % boundary
        return content_type, body.getvalue()



import json
from typing import Any, AsyncContextManager, Type

import aiofiles  # type: ignore
from starlette.background import BackgroundTask
from starlette.responses import FileResponse, HTMLResponse, JSONResponse, PlainTextResponse, Response

from pait.app.base import BaseAppHelper
from pait.core import Pait as _Pait
from pait.g import config
from pait.model import response

from ._app_helper import AppHelper

__all__ = ["pait"]


async def make_mock_response(pait_response: Type[response.PaitBaseResponseModel]) -> Response:
    if issubclass(pait_response, response.PaitJsonResponseModel):
        resp: Response = JSONResponse(json.loads(pait_response.get_example_value(json_encoder_cls=config.json_encoder)))
    elif issubclass(pait_response, response.PaitTextResponseModel):
        resp = PlainTextResponse(pait_response.get_example_value(), media_type=pait_response.media_type)
    elif issubclass(pait_response, response.PaitHtmlResponseModel):
        resp = HTMLResponse(pait_response.get_example_value(), media_type=pait_response.media_type)
    elif issubclass(pait_response, response.PaitFileResponseModel):
        named_temporary_file: AsyncContextManager = aiofiles.tempfile.NamedTemporaryFile()  # type: ignore
        f: Any = await named_temporary_file.__aenter__()
        await f.write(pait_response.get_example_value())
        await f.seek(0)

        async def close_file() -> None:
            await named_temporary_file.__aexit__(None, None, None)

        resp = FileResponse(f.name, media_type=pait_response.media_type, background=BackgroundTask(close_file))
    else:
        raise NotImplementedError(f"make_mock_response not support {pait_response}")
    resp.status_code = pait_response.status_code[0]
    if pait_response.header:
        resp.headers.update(pait_response.header)
    return resp


class Pait(_Pait):
    app_helper_class: "Type[BaseAppHelper]" = AppHelper
    # If you assign a value directly, it will become a bound function
    make_mock_response_fn: staticmethod = staticmethod(make_mock_response)


pait = Pait()



import inspect
import sys
from dataclasses import dataclass
from typing import Callable, Dict, ForwardRef, List, Optional, Type


@dataclass()
class FuncSig:
    """func inspect.Signature model"""

    func: Callable
    sig: "inspect.Signature"
    param_list: List["inspect.Parameter"]
    cbv_class: Optional[Type] = None


_func_sig_dict: Dict[Callable, FuncSig] = {}


def get_func_sig(func: Callable) -> FuncSig:
    """get func inspect.Signature model"""
    if func in _func_sig_dict:
        return _func_sig_dict[func]

    sig: inspect.Signature = inspect.signature(func)
    param_list: List[inspect.Parameter] = []
    for key in sig.parameters:
        if not (sig.parameters[key].annotation != sig.empty or sig.parameters[key].name == "self"):
            continue
        parameter: inspect.Parameter = sig.parameters[key]
        if isinstance(parameter.annotation, str):
            # get real type
            value: ForwardRef = ForwardRef(parameter.annotation, is_argument=False)
            setattr(
                parameter, "_annotation", value._evaluate(sys.modules[func.__module__].__dict__, None)  # type: ignore
            )
        param_list.append(parameter)

    # return_param = sig.return_annotation
    func_sig: FuncSig = FuncSig(func=func, sig=sig, param_list=param_list)
    _func_sig_dict[func] = func_sig
    return func_sig



from abc import ABC
from typing import Any, Dict, List, Optional

from tornado.web import Application, RequestHandler

from pait.api_doc.html import get_redoc_html as _get_redoc_html
from pait.api_doc.html import get_swagger_ui_html as _get_swagger_ui_html
from pait.api_doc.open_api import PaitOpenApi
from pait.field import Depends, Query
from pait.model.core import PaitCoreModel

from ._load_app import load_app
from ._pait import pait

__all__ = ["add_doc_route"]


def add_doc_route(
    app: Application,
    scheme: Optional[str] = None,
    prefix: str = "/",
    pin_code: str = "",
    title: str = "Pait Doc",
    open_api_tag_list: Optional[List[Dict[str, Any]]] = None,
) -> None:
    class NotFound(Exception):
        pass

    def _get_request_pin_code(r_pin_code: str = Query.i("", alias="pin_code")) -> Optional[str]:
        if pin_code:
            if r_pin_code != pin_code:
                raise NotFound
        return r_pin_code

    class BaseHandle(RequestHandler, ABC):
        def _handle_request_exception(self, exc: BaseException) -> None:
            self.set_status(404)
            self.write(
                (
                    "The requested URL was not found on the server. If you entered"
                    " the URL manually please check your spelling and try again."
                )
            )
            self.finish()

        def _get_open_json_url(self, r_pin_code: str) -> str:
            openapi_json_url: str = f"{'/'.join(self.request.path.split('/')[:-1])}/openapi.json"
            if r_pin_code:
                openapi_json_url += f"?pin_code={r_pin_code}"
            return openapi_json_url

    class GetRedocHtmlHandle(BaseHandle, ABC):
        @pait()
        async def get(self, r_pin_code: str = Depends.i(_get_request_pin_code)) -> None:
            self.write(_get_redoc_html(self._get_open_json_url(r_pin_code), title))

    class GetSwaggerUiHtmlHandle(BaseHandle, ABC):
        @pait()
        async def get(self, r_pin_code: str = Depends.i(_get_request_pin_code)) -> None:
            self.write(_get_swagger_ui_html(self._get_open_json_url(r_pin_code), title))

    class OpenApiHandle(BaseHandle, ABC):
        @pait(pre_depend_list=[_get_request_pin_code])
        async def get(self) -> None:
            pait_dict: Dict[str, PaitCoreModel] = load_app(self.application)
            _scheme: str = scheme or self.request.protocol
            pait_openapi: PaitOpenApi = PaitOpenApi(
                pait_dict,
                title=title,
                open_api_server_list=[{"url": f"{_scheme}://{self.request.host}", "description": ""}],
                open_api_tag_list=open_api_tag_list,
            )
            self.write(pait_openapi.open_api_dict)

    if not prefix.endswith("/"):
        prefix = prefix + "/"
    # Method 1
    # app.add_handlers(
    #     r".*$",
    #     [
    #         (r"{}redoc".format(prefix), GetRedocHtmlHandle),
    #         (r"{}swagger".format(prefix), GetSwaggerUiHtmlHandle),
    #         (r"{}openapi.json".format(prefix), OpenApiHandle),
    #     ]
    # )
    # Method 2
    # app.add_handlers(
    # app.default_router.add_rules(
    #     [
    #         (r"{}redoc".format(prefix), GetRedocHtmlHandle),
    #         (r"{}swagger".format(prefix), GetSwaggerUiHtmlHandle),
    #         (r"{}openapi.json".format(prefix), OpenApiHandle),
    #     ]
    # )

    # Method 3
    app.wildcard_router.add_rules(
        [
            (r"{}redoc".format(prefix), GetRedocHtmlHandle),
            (r"{}swagger".format(prefix), GetSwaggerUiHtmlHandle),
            (r"{}openapi.json".format(prefix), OpenApiHandle),
        ]
    )
    from tornado.web import AnyMatches, Rule, _ApplicationRouter

    app.default_router = _ApplicationRouter(app, [Rule(AnyMatches(), app.wildcard_router)])



from typing import Any, Type

import aiofiles  # type: ignore
from tornado.web import RequestHandler

from pait.app.base import BaseAppHelper
from pait.core import Pait as _Pait
from pait.g import config
from pait.model import response

from ._app_helper import AppHelper

__all__ = ["pait"]


async def make_mock_response(pait_response: Type[response.PaitBaseResponseModel]) -> Any:
    tornado_handle: RequestHandler = getattr(pait_response, "handle", None)
    if not tornado_handle:
        raise RuntimeError("Can not load Tornado handle")
    tornado_handle.set_status(pait_response.status_code[0])
    for key, value in pait_response.header.items():
        tornado_handle.set_header(key, value)
    tornado_handle.set_header("Content-Type", pait_response.media_type)
    if issubclass(pait_response, response.PaitJsonResponseModel):
        tornado_handle.write(pait_response.get_example_value(json_encoder_cls=config.json_encoder))
    elif issubclass(pait_response, response.PaitTextResponseModel) or issubclass(
        pait_response, response.PaitHtmlResponseModel
    ):
        tornado_handle.write(pait_response.get_example_value())
    elif issubclass(pait_response, response.PaitFileResponseModel):
        async with aiofiles.tempfile.NamedTemporaryFile() as f:  # type: ignore
            await f.write(pait_response.get_example_value())
            await f.seek(0)
            async for line in f:
                tornado_handle.write(line)
    else:
        raise NotImplementedError()


class Pait(_Pait):
    app_helper_class: "Type[BaseAppHelper]" = AppHelper
    # If you assign a value directly, it will become a bound function
    make_mock_response_fn: staticmethod = staticmethod(make_mock_response)


pait = Pait()



import logging
from typing import Any, List, Mapping, Tuple


class BaseAppHelper(object):
    """Provide a unified framework call interface for pait"""

    RequestType = type(None)
    FormType = type(None)
    FileType = type(None)
    HeaderType = type(None)
    app_name: str = "BaseAppHelper"

    def __init__(self, class_: Any, args: Tuple[Any, ...], kwargs: Mapping[str, Any]):
        """
        Extract the required data from the passed parameters,
        such as the self parameter in cvb mode, the request parameter in starletter
        """
        self.cbv_instance: Any = class_

        request = None
        new_args: List[Any] = []
        for param in args:
            if type(param) == self.RequestType:
                request = param
                # In cbv, request parameter will only appear after the self parameter
                break
            elif param == self.cbv_instance:
                pass
            else:
                # In cbv, parameter like self, request, {other param}
                # Now, not support other param
                logging.warning("Pait only support self and request args param")
                break
            new_args.append(param)

        self.request: Any = request
        self.request_args: List[Any] = new_args
        self.request_kwargs: Mapping[str, Any] = kwargs

    def cookie(self) -> Any:
        raise NotImplementedError

    def header(self) -> Any:
        raise NotImplementedError

    def path(self) -> Any:
        raise NotImplementedError

    def query(self) -> Any:
        raise NotImplementedError

    def body(self) -> Any:
        raise NotImplementedError

    def file(self) -> Any:
        raise NotImplementedError

    def form(self) -> Any:
        raise NotImplementedError

    def multiform(self) -> Any:
        raise NotImplementedError

    def multiquery(self) -> Any:
        raise NotImplementedError

    def check_request_type(self, value: Any) -> bool:
        return value is self.RequestType

    def check_file_type(self, value: Any) -> bool:
        return value is self.FileType

    def check_form_type(self, value: Any) -> bool:
        return value is self.FormType

    def check_header_type(self, value: Any) -> bool:
        return value is self.HeaderType



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pait',
 'pait.api_doc',
 'pait.api_doc.html',
 'pait.app',
 'pait.app.base',
 'pait.app.flask',
 'pait.app.sanic',
 'pait.app.starlette',
 'pait.app.tornado',
 'pait.model',
 'pait.util']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0', 'pydantic>=1.7.3,<2.0.0']

setup_kwargs = {
    'name': 'pait',
    'version': '0.7.2',
    'description': 'Pait is a Python api tool. Pait enables your Python web framework to have type checking, parameter type conversion, interface document generation and can display your documents through Redoc or Swagger (power by inspect, pydantic)',
    'long_description': '# pait\nPait is an api tool that can be used in any python web framework (currently only `flask`, `starlette`, `sanic`, `tornado` are supported, other frameworks will be supported once Pait is stable).\n\nPait allows the Python Web framework to have functions such as parameter type checking, type conversion (depending on Pydantic and inspect), and document output.\n\n> Note:\n>\n> mypy check 100%\n>\n> test coverage 95%+ (exclude api_doc)\n>\n> python version >= 3.7 (support postponed annotations)\n>\n> The function is being expanded... the documentation may not be perfect\n>\n> The following code does not specify, all default to use the `starlette` framework.\n>\n> There is no test case for the document output function, and the function is still being improved\n\n[中文文档](https://github.com/so1n/pait/blob/master/README_ZH.md)\n# Feature\n - [x] Parameter checksum automatic conversion (parameter check depends on `Pydantic`)\n - [x] Parameter dependency verification\n - [x] Automatically generate openapi files\n - [x] Support swagger, redoc routing\n - [x] return mock response\n - [x] TestClient support, support response result verification\n - [ ] Support more types of http requests (currently only supports RESTful api)\n - [ ] Combine faker to provide better mock response\n - [ ] Local api document management\n# Installation\n```Bash\npip install pait\n```\n# Usage\n## 1.type checking and parameter type conversion\n### 1.1.Use in route handle\nA simple starlette route handler example:\n```Python\nimport uvicorn\nfrom starlette.applications import Starlette\nfrom starlette.requests import Request\nfrom starlette.responses import JSONResponse\nfrom starlette.routing import Route\n\n\nasync def demo_post(request: Request) -> JSONResponse:\n    body_dict: dict = await request.json()\n    uid: int = body_dict.get(\'uid\', 0)\n    user_name: str = body_dict.get(\'user_name\', "")\n    # The following code is only for demonstration, in general, we do some wrapping\n    if not uid:\n        raise ValueError(\'xxx\')\n    if type(uid) != int:\n        raise TypeError(\'xxxx\')\n    if 10 <= uid <= 1000:\n        raise ValueError(\'xxx\')\n\n    if not user_name:\n        raise ValueError(\'xxx\')\n    if type(user_name) != str:\n        raise TypeError(\'xxxx\')\n    if 2 <= len(user_name) <= 4:\n        raise ValueError(\'xxx\')\n\n    return JSONResponse(\n        {\n            \'result\': {\n                \'uid\': body_dict[\'uid\'],\n                \'user_name\': body_dict[\'user_name\']\n            }\n        }\n    )\n\n\napp = Starlette(\n    routes=[\n        Route(\'/api\', demo_post, methods=[\'POST\']),\n    ]\n)\n\n\nuvicorn.run(app)\n```\nuse pait in starletter route handler:\n\n```Python\nimport uvicorn  # type: ignore\nfrom starlette.applications import Starlette\nfrom starlette.responses import JSONResponse\nfrom starlette.routing import Route\n\nfrom pait.app.starlette import pait\nfrom pait.field import Body\n\n\n@pait()\nasync def demo_post(\n    uid: int = Body.i(description="user id", gt=10, lt=1000),\n    user_name: str = Body.i(description="user name", min_length=2, max_length=4)\n) -> JSONResponse:\n    return JSONResponse({\'result\': {\'uid\': uid, \'user_name\': user_name}})\n\n\napp = Starlette(\n    routes=[\n        Route(\'/api\', demo_post, methods=[\'POST\']),\n    ]\n)\n\nuvicorn.run(app)\n```\nIt can be seen from the above that the code has been refined a lot. All of this is the role of the `pait` decorator. It finds out how to get the value through the function signature, what is the type of the value, and what is the key corresponding to the value. After assembly, it is handed over to `Pydantic` for verification and conversion, and then returned to the corresponding parameters of the routing function according to the function signature.\n\nfor example param: uid\n```Python3\nfrom pait.field import Body\n\nuid: int = Body.i(description="user id", gt=10, lt=1000)\n```\n`pait` will be split into the following parts:\n\n```\n<key>: <type> = <request data>\n```\nThe key is the parameter name, type is the parameter type, and request data is the other description of the parameter. For example, body represents the data of the request body, gt is the minimum parameter, and lt is the maximum parameter.\n\n\n\nHere is just a simple demo, because we write the model can be reused, so you can save a lot of development time, the above parameters are only used to a way to write, the following will introduce pait support for the two ways to write and use.\n\n### 1.2.Parameter expression supported by pait\npait in order to facilitate the use of users, support a variety of writing methods (mainly the difference between TypeHints)\n- TypeHints is BaseModel, mainly used for parameters from multiple `Field`, and want to reuse model:\n\n    BaseModel can be used only for args parameters:\n    ```Python\n    from pait.app.starlette import pait\n    from pait.field import Body, Header\n    from pydantic import BaseModel\n    ```\n\n\n    class TestModel(BaseModel):\n        uid: int = Body.i()\n        content_type: str = Header.i(default=\'Content-Type\')\n\n\n    @pait()\n    async def test(model: BaseModel):\n        return {\'result\': model.dict()}\n    ```\n- TypeHints is Pydantic.BaseModel, mainly used for parameters are derived from the same `Field`, and want to take the model:\n\n    BaseModel can only be used with kwargs parameters, and the type hints of the parameters must be a class that inherits from `pydantic.BaseModel`, using the example:\n    ````Python\n    from pydantic import BaseModel\n\n    from pait.app.starlette import pait\n    from pait.field import Body\n\n\n    class TestModel(BaseModel):\n        uid: int\n        user_name: str\n\n\n    @pait()\n    async def test(model: BaseModel = Body.i()):\n        return {\'result\': model.dict()}\n    ````\n- When TypeHints is not one of the above two cases:\n\n    can only be used for kwargs parameters and type hints are not the above two cases, if the value is rarely reused, or if you do not want to create a Model, you can consider this approach\n    ```Python\n    from pait.app.starlette import pait\n    from pait.field import Body\n    ```\n\n\n    @pait()\n    async def test(uid: int = Body.i(), user_name: str = Body.i()):\n        return {\'result\': {\'uid\': uid, \'user_name\': user_name}}\n    ```\n### 1.3.Field\nField will help pait know how to get data from request.\nBefore introducing the function of Field, let’s take a look at the following example. `pait` will obtain the body data of the request according to Field.Body, and obtain the value with the parameter named key. Finally, the parameter is verified and assigned to the uid.\n\n> Note: Use Field.Body() directly, `mypy` will check that the type does not match, then just change to Field.Body.i() to solve the problem.\n```Python\nfrom pait.app.starlette import pait\nfrom pait.field import Body\n\n\n@pait()\nasync def demo_post(\n    # get uid from request body data\n    uid: int = Body.i()\n) -> None:\n    pass\n```\n`Pait` also supports other functions when performing parameter verification and conversion, all of which are supported by the parameters of `<request data>`:\n\n- default: the function of providing default values, if the request parameter does not have the value of this parameter , The value is used by default\n\n- alias: Since `Content-Type` cannot be named in Python variables, it can only be named with `content_type` according to the naming convention of `Python`, and `content_type` cannot get the value directly from the header, so it can Set alias to `Content-Type`, so that `Pait` can get the value of `Content-Type` in the Header and assign it to the `content_type` variable.\n\n- raw_return: If the value is True, `Pait` will not use the parameter name to get the data, but directly assign the entire data to the corresponding parameter.\n\n```Python\nfrom pait.app.starlette import pait\nfrom pait.field import Body, Header\n\n\n@pait()\nasync def demo_post(\n    # get uid from request body data\n    uid: int = Body.i(default=100),\n    # get Content-Type from header\n    content_type: str = Header.i(alias=\'Content-Type\'),\n    header_dict: str = Header.i(raw_return=True)\n\n):\n    pass\n```\nThe above only demonstrates the Body and Header of the field, but there are other fields as well::\n- Field.Body   Get the json data of the current request\n- Field.Cookie Get the cookie data of the current request\n- Field.File   Get the file data of the current request, depending on the web framework will return different file object types\n- Field.Form   Get the form data of the current request, if there are multiple duplicate keys, only the first one will be returned\n- Field.Header Get the header data of the current request\n- Field.Path   Get the path data of the current request (e.g. /api/{version}/test, you can get the version data)\n- Field.Query  Get the url parameters of the current request and the corresponding data, if there are multiple duplicate keys, only the first one will be returned\n- Field.MultiQuery Get the url parameter data of the current request, and return the list corresponding to the key\n- Field.MultiForm Get the form data of the current request, return the list corresponding to the key\n\nAll the fields above are inherited from `pydantic.fields.FieldInfo`, most of the parameters here are for api documentation, see for specific usage[pydantic doc](https://pydantic-docs.helpmanual.io/usage/schema/#field-customisation)\n\n\nIn addition there is a field named Depends, he inherits from `object`, he provides the function of dependency injection, he only supports one parameter and the type of function, and the function\'s parameters are written in the same way as the routing function, the following is an example of the use of Depends, through Depends, you can reuse in each function to get the token function:\n\n```Python\nfrom pait.app.starlette import pait\nfrom pait.field import Body, Depends\n\n\ndef demo_depend(uid: str = Body.i(), password: str = Body.i()) -> str:\n    # fake db\n    token: str = db.get_token(uid, password)\n    return token\n\n\n@pait()\nasync def test_depend(token: str = Depends.i(demo_depend)) -> dict:\n    return {\'token\': token}\n```\n\n### 1.4.requests object\nAfter using `Pait`, the proportion of the number of times the requests object is used will decrease, so `pait` does not return the requests object. If you need the requests object, you can fill in the parameters like `requests: Requests` (you need to use the TypeHints format) , You can get the requests object corresponding to the web framework\n```Python\nfrom starlette.requests import Request\n\nfrom pait.app.starlette import pait\nfrom pait.field import Body\n\n\n@pait()\nasync def demo_post(\n        request: Request,\n        # get uid from request body data\n        uid: int = Body.i()\n) -> None:\n    pass\n```\n\n### 1.5.Exception\n#### 1.5.1Exception Handling\nPait will leave the exception to the user to handle it. Under normal circumstances, pait will only throw the exception of `pydantic` and `PaitBaseException`. The user needs to catch the exception and handle it by himself, for example:\n```Python\nfrom starlette.applications import Starlette\nfrom starlette.requests import Request\nfrom starlette.responses import Response\n\nfrom pait.exceptions import PaitBaseException\nfrom pydantic import ValidationError\n\nasync def api_exception(request: Request, exc: Exception) -> None:\n    """\n    Handle exception code\n    """\n    if isinstance(exc, PaitBaseException):\n        pass\n    elif isinstance(exc, ValidationError):\n        pass\n    else:\n        pass\n\nAPP = Starlette()\nAPP.add_exception_handler(PaitBaseException, api_exception)\nAPP.add_exception_handler(ValidationError, api_exception)\n```\n#### 1.5.2Error Tip\nWhen you use `pait` incorrectly, `pait` will indicate in the exception the file path and line number of the function.\n```Bash\n  File "/home/so1n/github/pait/pait/func_param_handle.py", line 101, in set_value_to_kwargs_param\n    f\'File "{inspect.getfile(func_sig.func)}",\'\nPaitBaseException: \'File "/home/so1n/github/pait/example/starlette_example.py", line 28, in demo_post\\n kwargs param:content_type: <class \\\'str\\\'> = Header(key=None, default=None) not found value, try use Header(key={key name})\'\n```\nIf you need more information, can set the log level to debug to get more detailed information\n```Python\nDEBUG:root:\nasync def demo_post(\n    ...\n    content_type: <class \'str\'> = Header(key=None, default=None) <-- error\n    ...\n):\n    pass\n```\n## 2.Document Generation\n\n`pait` will automatically capture the request parameters and url, method and other information of the routing function.\nIn addition, it also supports labeling some relevant information. These labels will only be loaded into the memory when the Python program starts running, and will not affect the performance of the request, as in the following example:\n\n```Python\nfrom pait.app.starlette import pait\nfrom pait.model.status import PaitStatus\n\nfrom example.param_verify.model import UserSuccessRespModel, FailRespModel\n\n\n@pait(\n  author=("so1n",),\n  group="user",\n  status=PaitStatus.release,\n  tag=("user", "post"),\n  response_model_list=[UserSuccessRespModel, FailRespModel],\n)\ndef demo() -> None:\n  pass\n```\nParam:\n- author: List of authors who wrote the interface\n- group: The group to which the interface belongs (This option is currently not used for openapi)\n- status: The status of the interface, currently only supports several states of `PaitStatus` (This option will only be used for openapi and marked as deprecated if it is offline)\n  - default status:\n    - undefined: undefined\n  - in development:\n    - design: Interface design\n    - dev: Under development and testing\n  - Development completed:\n    - integration: integration test\n    - complete: development completed\n    - test: testing\n  - online:\n    - release: online\n  - offline:\n    - abnormal: The interface is abnormal and needs to be offline\n    - maintenance: In maintenance\n    - archive: archive\n    - abandoned: abandoned\n- tag: interface tag\n- response_model_list: return data, Need to inherit from `pait.model.PaitResponseModel`, Since `pait` is an extension of the web framework and will not modify the code of the framework, this parameter will not be used for ordinary request judgment (nor should it be used in the production environment). It is currently only used for document generation, mock response generation and TestClient verification.\n\n### 2.1.openapi\n#### 2.1.1openapi doc output\nCurrently pait supports most of the functions of openapi, a few unrealized features will be gradually improved through iterations (response-related more complex)\n\nThe openapi module of pait supports the following parameters (more parameters will be provided in the next version):\n- title: openapi\'s title\n- open_api_info: openapi\'s info param\n- open_api_tag_list: related description of openapi tag\n- open_api_server_list: openapi server list\n- type_: The type of output, optionally json and yaml\n- filename: Output file name, or if empty, output to terminal\n\nThe following is the sample code output from the openapi documentation (modified by the 1.1 code). See [Example code](https://github.com/so1n/pait/tree/master/example/api_doc) and [doc example](https://github.com/so1n/pait/blob/master/example/api_doc/example_doc)\n```Python\nimport uvicorn\n\nfrom starlette.applications import Starlette\nfrom starlette.routing import Route\nfrom starlette.responses import JSONResponse\n\nfrom pait.field import Body\nfrom pait.app.starlette import pait\nfrom pydantic import (\n    BaseModel,\n    conint,\n    constr,\n)\n\n\n# Create a Model based on Pydantic.BaseModel\nclass PydanticModel(BaseModel):\n    uid: conint(gt=10, lt=1000)  # Whether the auto-check type is int, and whether it is greater than or equal to 10 and less than or equal to 1000\n    user_name: constr(min_length=2, max_length=4)  # Whether the auto-check type is str, and whether the length is greater than or equal to 2, less than or equal to 4\n\n\n\n# Decorating functions with the pait decorator\n@pait()\nasync def demo_post(\n    # pait through the Body () to know the current need to get the value of the body from the request, and assign the value to the model,\n    # and the structure of the model is the above PydanticModel, he will be based on our definition of the field automatically get the value and conversion and judgment\n    model: PydanticModel = Body.i()\n):\n    # Get the corresponding value to return\n    return JSONResponse({\'result\': model.dict()})\n\n\napp = Starlette(\n    routes=[\n        Route(\'/api\', demo_post, methods=[\'POST\']),\n    ]\n)\n\nuvicorn.run(app)\n# --------------------\n\nfrom pait.app.starlette import load_app\nfrom pait.api_doc.open_api import PaitOpenApi\n\n\n# Extracting routing information to pait\'s data module\npait_dict = load_app(app)\n# Generate openapi for routing based on data from the data module\nPaitOpenApi(pait_dict)\n```\n#### 2.1.2.OpenApi Route\n`Pait` currently supports openapi.json routing, and also supports page display of `Redoc` and `Swagger`, and these only need to call the `add_doc_route` function to add three routes to the `app` instance:\n- /openapi.json\n- /redoc\n- /swagger\nIf you want to define a prefix, such as /doc/openapi.json, just pass in /doc through the prefix parameter. Specific examples are as follows:\n```Python\nimport uvicorn  # type: ignore\nfrom pydantic import BaseModel, Field\nfrom starlette.applications import Starlette\nfrom starlette.responses import JSONResponse\nfrom starlette.routing import Route\n\nfrom pait.app.starlette import pait\nfrom pait.field import Body\n\n\n# Create a Model based on Pydantic.BaseModel\nclass UserModel(BaseModel):\n    # Whether the auto-check type is int, and whether it is greater than or equal to 10 and less than or equal to 1000\n    uid: int = Field(description="user id", gt=10, lt=1000)\n    # Whether the auto-check type is str, and whether the length is greater than or equal to 2, less than or equal to 4\n    user_name: str = Field(description="user name", min_length=2, max_length=4)\n\n\n# Decorating functions with the pait decorator\n@pait()\nasync def demo_post(\n    # pait through the Body () to know the current need to get the value of the body from the request, and assign the value to the model,\n    # and the structure of the model is the above PydanticModel, he will be based on our definition of the field automatically get the value and conversion and judgment\n    model: UserModel = Body.i()\n) -> JSONResponse:\n    # Get the corresponding value to return\n    return JSONResponse({\'result\': model.dict()})\n\n\napp = Starlette(\n    routes=[\n        Route(\'/api\', demo_post, methods=[\'POST\']),\n    ]\n)\n# Inject the route into the app\nadd_doc_route(app)\n# Inject the route into the app, and prefix it with /doc\nadd_doc_route(app, prefix=\'/doc\')\n```\n### 2.2.Other doc output\n> Note: The function is being improved...\n\nIn addition to parameter verification and conversion, pait also provides the ability to output api documentation, which can be configured with simple parameters to output perfect documentation.\n\nNote: Currently only md documents and openapi documents for json and yaml are supported for output.For the output of md, see\n[doc example](https://github.com/so1n/pait/blob/master/example/api_doc/example_doc)\n\n\n## 3.Implicit and explicit introduction\n`pait` provides support for multiple frameworks. If only one of the frameworks is installed in a project, then you can use implicit import:\n```Python3\nfrom pait.app import add_doc_route, load_app, pait\n\n```\nHowever, if multiple frameworks are installed at the same time, the above introduction will throw an error. It is recommended to use explicit introduction, such as:\n```Python3\nfrom pait.app.starlette import add_doc_route, load_app, pait\n\n```\n## 4.config, data and load_app\n- data\n\n  Since `pait` provides functional support through a decorator, all data is injected into meta data when the compiler is started, providing support for subsequent document generation and other functions.\n- load_app\n\n  There are a lot of routing function information in meta data, but it lacks key parameters such as `url`, `method`, etc.\nSo you also need to use load_app to bind the relevant parameters to the routing function data decorated by the `pait` decorator in the meta data. The method of use is very simple, but remember that you must register all routes before calling:\n```Python3\n  from starlette.applications import Starlette\n\n  from pait.app.starlette import load_app\n\n  app: Starlette = Starlette()\n  # error\n  load_app(app)\n  # --------\n  # app.add_route\n  # --------\n\n  # success\n  load_app(app)\n  ```\n- config\nconfig can provide some configuration support for `pait`, it needs to be initialized as soon as possible. The best initialization position is to initialize before app initialization, and only one initialization is allowed during the entire runtime.\n  ```Python\n  from starlette.applications import Starlette\n\n  from pait.app.starlette import load_app\n  from pait.g import config\n\n  config.init_config(author="so1n")\n  app: Starlette = Starlette()\n  # --------\n  # app.add_route\n  # --------\n  load_app(app)\n  ```\n\nParameter introduction:\n- author: The global default API author, if the author parameter in `@pait` is empty, it will call `config.author` by default.\n- status: The global default API status, if the status in `@pait` is empty, it will be called by default to `config.status`\n- enable_mock_response: Decide whether this run will return a normal response or a mock response based on `response_model`\n- enable_mock_response_filter_fn: Multiple `response_model` are supported by default, and the mock response only takes the first `response_model` by default. If you feel that this does not meet the `response_model` you want, you can configure this function to return the results you want\n- block_http_method_set: Some web frameworks will automatically help add some routing functions to request methods such as `HEAD`. `pait` cannot recognize which are added by the framework and which are added by the user. Users can block some `methods` through this parameter\n- default_response_model_list: When designing some API interfaces, there are usually some default exception responses, and repeated configuration is very troublesome. can apply to the global by configuring this parameter\n- json_type_default_value_dict: Configure the default value of the json type\n## 5.TestClientHelper\n`pait` encapsulates a corresponding `TestCLientHelper` class for each framework, through which test cases can be written more conveniently, and the result data structure can be compared with `response_model` for verification. [starlette example](https://github.com/so1n/pait/blob/master/tests/test_app/test_starlette.py#L80)\n\nParameter Description:\n  - client: The test client corresponding to the framework\n  - func: Corresponding to the routing function decorated by `pait`\n  - pait_dict: `pait` meta data, if it is empty, it will be automatically generated internally\n  - body_dict: Requested json data\n  - cookie_dict: Requested cookie data\n  - file_dict: Requested file data\n  - form_dict: Requested form data\n  - header_dict: Requested header data\n  - path_dict: Requested path data\n  - query_dict: Requested query data\n\n\nDescription of response_model generation results:\nIf response_model has a `default` value, it will be directly referenced to the `default` value, otherwise the default value of the value type is used (configurable through config)\n## 6.How to used in other web framework?\nIf the web framework is not supported, which you are using.\nCan be modified sync web framework according to [pait.app.flask](https://github.com/so1n/pait/blob/master/pait/app/flask.py)\n\nCan be modified async web framework according to [pait.app.starlette](https://github.com/so1n/pait/blob/master/pait/app/starlette.py)\n## 7.IDE Support\nWhile pydantic will work well with any IDE out of the box.\n- [PyCharm plugin](https://pydantic-docs.helpmanual.io/pycharm_plugin/)\n- [Mypy plugin](https://pydantic-docs.helpmanual.io/mypy_plugin/)\n\n## 8.Full example\nFor more complete examples, please refer to [example](https://github.com/so1n/pait/tree/master/example)\n',
    'author': 'So1n',
    'author_email': 'so1n897046026@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/so1n/pait',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import inspect
from functools import wraps
from typing import Any, Callable, Dict, List, Optional, Tuple, Type

from pydantic import BaseConfig

from pait.app.base import BaseAppHelper
from pait.g import config, pait_data
from pait.model.core import PaitCoreModel
from pait.model.response import PaitBaseResponseModel
from pait.model.status import PaitStatus
from pait.model.util import sync_config_data_to_pait_core_model
from pait.param_handle import AsyncParamHandler, ParamHandler
from pait.util import get_func_sig


class Pait(object):
    app_helper_class: "Type[BaseAppHelper]"
    make_mock_response_fn: staticmethod  # real type hints: Callable[[Type[PaitBaseResponseModel]], Any]]

    def __init__(
        self,
        make_mock_response_fn: Optional[Callable[[Type[PaitBaseResponseModel]], Any]] = None,
        enable_mock_response: bool = False,
        pydantic_model_config: Optional[Type[BaseConfig]] = None,
        # param check
        pre_depend_list: Optional[List[Callable]] = None,
        at_most_one_of_list: Optional[List[List[str]]] = None,
        required_by: Optional[Dict[str, List[str]]] = None,
        # doc
        author: Optional[Tuple[str, ...]] = None,
        desc: Optional[str] = None,
        summary: Optional[str] = None,
        name: Optional[str] = None,
        status: Optional[PaitStatus] = None,
        group: Optional[str] = None,
        tag: Optional[Tuple[str, ...]] = None,
        response_model_list: Optional[List[Type[PaitBaseResponseModel]]] = None,
    ):

        check_cls_param_list: List[str] = ["app_helper_class", "make_mock_response_fn"]
        for cls_param in check_cls_param_list:
            if not getattr(self, cls_param, None):
                raise ValueError(
                    f"Please specify the value of the {cls_param} parameter, you can refer to `pait.app.xxx`"
                )
        if not isinstance(self.app_helper_class, type):
            raise TypeError(f"{self.app_helper_class} must be class")
        if not issubclass(self.app_helper_class, BaseAppHelper):
            raise TypeError(f"{self.app_helper_class} must sub from {BaseAppHelper.__class__.__name__}")
        if getattr(self.make_mock_response_fn, "__self__", None):
            raise TypeError(
                "Set error make_mock_response_fn, should use `staticmethod` func. "
                "like `make_mock_response_fn = staticmethod(make_mock_response)`"
            )

        self._make_mock_response_fn: Callable[[Type[PaitBaseResponseModel]], Any] = (
            make_mock_response_fn or self.make_mock_response_fn
        )
        self._enable_mock_response: bool = enable_mock_response
        self._pydantic_model_config: Optional[Type[BaseConfig]] = pydantic_model_config
        # param check
        self._pre_depend_list: Optional[List[Callable]] = pre_depend_list
        self._at_most_one_of_list: Optional[List[List[str]]] = at_most_one_of_list
        self._required_by: Optional[Dict[str, List[str]]] = required_by
        # doc
        self._author: Optional[Tuple[str, ...]] = author
        self._desc: Optional[str] = desc
        self._summary: Optional[str] = summary
        self._name: Optional[str] = name
        self._status: Optional[PaitStatus] = status
        self._group: Optional[str] = group
        self._tag: Optional[Tuple[str, ...]] = tag
        self._response_model_list: Optional[List[Type[PaitBaseResponseModel]]] = response_model_list

    def __call__(
        self,
        make_mock_response_fn: Callable[[Type[PaitBaseResponseModel]], Any] = None,
        enable_mock_response: bool = False,
        pydantic_model_config: Optional[Type[BaseConfig]] = None,
        # param check
        pre_depend_list: Optional[List[Callable]] = None,
        at_most_one_of_list: Optional[List[List[str]]] = None,
        required_by: Optional[Dict[str, List[str]]] = None,
        # doc
        author: Optional[Tuple[str, ...]] = None,
        desc: Optional[str] = None,
        summary: Optional[str] = None,
        name: Optional[str] = None,
        status: Optional[PaitStatus] = None,
        group: Optional[str] = None,
        tag: Optional[Tuple[str, ...]] = None,
        response_model_list: Optional[List[Type[PaitBaseResponseModel]]] = None,
    ) -> Callable:
        app_name: str = self.app_helper_class.app_name
        enable_mock_response = enable_mock_response or self._enable_mock_response
        pydantic_model_config = pydantic_model_config or self._pydantic_model_config
        pre_depend_list = pre_depend_list or self._pre_depend_list
        at_most_one_of_list = at_most_one_of_list or self._at_most_one_of_list
        required_by = required_by or self._required_by
        author = author or self._author
        desc = desc or self._desc
        summary = summary or self._summary
        name = name or self._name
        group = group or self._group
        tag = tag or self._tag
        response_model_list = response_model_list or self._response_model_list

        def wrapper(func: Callable) -> Callable:
            # Pre-parsing function signatures
            get_func_sig(func)

            # gen pait core model and register to pait data
            pait_core_model: PaitCoreModel = PaitCoreModel(
                author=author,
                app_helper_class=self.app_helper_class,
                make_mock_response_fn=make_mock_response_fn or self._make_mock_response_fn,
                desc=desc,
                summary=summary,
                func=func,
                func_name=name,
                status=status,
                group=group,
                tag=tag,
                response_model_list=response_model_list,
                pre_depend_list=pre_depend_list,
                pydantic_model_config=pydantic_model_config,
            )
            sync_config_data_to_pait_core_model(config, pait_core_model, enable_mock_response=enable_mock_response)
            pait_data.register(app_name, pait_core_model)

            if inspect.iscoroutinefunction(func):

                @wraps(func)
                async def dispatch(*args: Any, **kwargs: Any) -> Callable:
                    async with AsyncParamHandler(
                        self.app_helper_class,
                        func,
                        pait_core_model.pydantic_model_config,
                        at_most_one_of_list=at_most_one_of_list,
                        pre_depend_list=pre_depend_list,
                        required_by=required_by,
                        args=args,
                        kwargs=kwargs,
                    ) as p:
                        return await pait_core_model.pait_func(*p.args, **p.kwargs)

                return dispatch
            else:

                @wraps(func)
                def dispatch(*args: Any, **kwargs: Any) -> Callable:
                    with ParamHandler(
                        self.app_helper_class,
                        func,
                        pait_core_model.pydantic_model_config,
                        at_most_one_of_list=at_most_one_of_list,
                        pre_depend_list=pre_depend_list,
                        required_by=required_by,
                        args=args,
                        kwargs=kwargs,
                    ) as p:
                        return pait_core_model.pait_func(*p.args, **p.kwargs)

                return dispatch

        return wrapper
